#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test deals with the surrogate modeling of a Borehole function.

You will see how to:
    Check the quality of your regression model
    Perform sensitivity analysis via Sobol Indices

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Sep 12 2022

"""

import numpy as np
import joblib

# import bayesvalidrox
# Add BayesValidRox path
import sys
sys.path.append("../../src/bayesvalidrox/")

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
#from bayesvalidrox.surrogate_models.meta_model_engine import MetaModelEngine
from bayesvalidrox.post_processing.post_processing import PostProcessing
from bayesvalidrox.bayes_inference.bayes_inference import BayesInference
from bayesvalidrox.bayes_inference.discrepancy import Discrepancy
from bayesvalidrox.surrogate_models.engine import Engine

import matplotlib
matplotlib.use('agg')


# TODO: this is a test in hopes of removing multiprocessing issues
from multiprocessing import set_start_method
if __name__ == "__main__":

    #set_start_method('fork', force = True)

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    # Define model options
    Model.link_type = 'Function'
    Model.py_file = 'borehole'
    Model.name = 'borehole'

    Model.Output.names = ['flow rate']#' [m$^3$/yr]']

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    Inputs = Input()

    # borehole radius
    Inputs.add_marginals()
    Inputs.Marginals[0].name = '$r_w$'
    Inputs.Marginals[0].dist_type = 'normal'
    Inputs.Marginals[0].parameters = [0.10, 0.0161812]

    # borehole length
    Inputs.add_marginals()
    Inputs.Marginals[1].name = '$L$'
    Inputs.Marginals[1].dist_type = 'unif'
    Inputs.Marginals[1].parameters = [1120, 1680]

    # borehole hydraulic conductivity
    Inputs.add_marginals()
    Inputs.Marginals[2].name = '$K_w$'
    Inputs.Marginals[2].dist_type = 'unif'
    Inputs.Marginals[2].parameters = [9855, 12045]

    # transmissivity of upper aquifer
    Inputs.add_marginals()
    Inputs.Marginals[3].name = '$T_u$'
    Inputs.Marginals[3].dist_type = 'unif'
    Inputs.Marginals[3].parameters = [63070, 115600]

    # transmissivity of lower aquifer
    Inputs.add_marginals()
    Inputs.Marginals[4].name = '$T_l$'
    Inputs.Marginals[4].dist_type = 'unif'
    Inputs.Marginals[4].parameters = [63.1, 116]

    # potentiometric head of upper aquifer
    Inputs.add_marginals()
    Inputs.Marginals[5].name = '$H_u$'
    Inputs.Marginals[5].dist_type = 'unif'
    Inputs.Marginals[5].parameters = [990, 1110]

    # potentiometric head of lower aquifer
    Inputs.add_marginals()
    Inputs.Marginals[6].name = '$H_l$'
    Inputs.Marginals[6].dist_type = 'unif'
    Inputs.Marginals[6].parameters = [700, 820]

    # radius of influence
    Inputs.add_marginals()
    Inputs.Marginals[7].name = '$r$'
    Inputs.Marginals[7].dist_type = 'lognorm'
    Inputs.Marginals[7].parameters = [7.71, 1.0056]

    # =====================================================
    # ======  POLYNOMIAL CHAOS EXPANSION METAMODELS  ======
    # =====================================================
    MetaModelOpts = MetaModel(Inputs, Model)

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)BCS: Bayesian Compressive Sensing
    # 7)OMP: Orthogonal Matching Pursuit
    # 8)VBL: Variational Bayesian Learning
    # 9)EBL: Emperical Bayesian Learning
    MetaModelOpts.pce_reg_method = 'OMP'

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    MetaModelOpts.pce_deg = 5

    # q-quasi-norm 0<q<1 (default=1)
    MetaModelOpts.pce_q_norm = 1.0

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    ExpDesign = ExpDesigns(Inputs)

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    ExpDesign.method = 'sequential'
    ExpDesign.n_init_samples = 50

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) korobov
    # 7) chebyshev(FT) 8) grid(FT) 9) nested_grid(FT) 10)user
    ExpDesign.sampling_method = 'latin_hypercube'

    # Provide the experimental design object with a hdf5 file
    # MetaModelOpts.ExpDesign.hdf5_file = 'ExpDesign_borehole.hdf5'

    # Sequential experimental design (needed only for sequential ExpDesign)
    ExpDesign.n_new_samples = 1
    ExpDesign.n_max_samples = 300
    ExpDesign.mod_LOO_threshold = 1e-16

    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # 1) None 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    ExpDesign.tradeoff_scheme = None
    # MetaModelOpts.ExpDesign.n_replication = 50
    # -------- Exploration ------
    # 1)'Voronoi' 2)'random' 3)'latin_hypercube' 4)'dual annealing'
    ExpDesign.explore_method = 'latin_hypercube'

    # Use when 'dual annealing' chosen
    ExpDesign.max_func_itr = 200

    # Use when 'Voronoi' or 'random' or 'latin_hypercube' chosen
    ExpDesign.n_canddidate = 5000
    ExpDesign.n_cand_groups = 4

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'VarOptDesign' 3)'alphabetic' 4)'Space-filling'
    ExpDesign.exploit_method = 'Space-filling'

    # BayesOptDesign -> when data is available
    # 1)DKL (Kullback-Leibler Divergence) 2)DPP (D-Posterior-percision)
    # 3)APP (A-Posterior-percision)
    # MetaModelOpts.ExpDesign.util_func = 'DKL'

    # VarBasedOptDesign -> when data is not available
    # Only with Vornoi >>> 1)Entropy 2)EIGF, 3)ALM, 4)LOOCV
    ExpDesign.util_func = 'ALM'

    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    # MetaModelOpts.ExpDesign.util_func = 'D-Opt'

    valid_samples = np.load("data/valid_samples.npy")
    #valid_model_runs = {'flow rate [m$^3$/yr]': np.load("data/valid_outputs.npy")}
    valid_model_runs = {'flow rate': np.load("data/valid_outputs.npy")}
    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Use MetaModelEngine for sequential experimental design
    
    MetaModelOpts.ExpDesign = ExpDesign
    engine = Engine(MetaModelOpts, Model, ExpDesign)
    engine.start_engine()
#    engine.train_sequential()
    engine.train_sequential(parallel = True)

    # Save PCE models
    with open(f'PCEModel_{Model.name}.pkl', 'wb') as output:
        joblib.dump(engine.MetaModel, output, 2)

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(engine)

    # Plot to check validation visually.
    PostPCE.valid_metamodel(n_samples=200)

    # Check the quality of your regression model
    PostPCE.check_reg_quality()

    # PostPCE.eval_PCEmodel_3D()
    # Compute and print RMSE error
    PostPCE.check_accuracy(n_samples=3000)

    # Plot the evolution of the KLD,BME, and Modified LOOCV error
    if MetaModelOpts.ExpDesign.method == 'sequential':
        PostPCE.plot_seq_design_diagnostics()

    # Plot the sobol indices
    total_sobol = PostPCE.sobol_indices(plot_type='bar')
