#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This example shows comparisons of three regression options for the aPC in 
bayesvalidrox: OLS, FastARD and OLS with constraints.
The trained surrogates are compared with respect to the here defined error
measure and their computational cost for different numbers of training samples.
"""

import numpy as np
import matplotlib.pyplot as plt
import random

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.engine import Engine
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
import bayesvalidrox.surrogate_models.surrogate_models as sm

from convergence_test import convergence_test
from testfunction import testfunction

def error(mean, std, true):
    """
    Error function to evaluate the metamodels
    Returns a dictionary of measures.
    Here:
        'L2': standard L2 error

    Parameters
    ----------
    mean : dict
        Mean-part of surrogate evaluation
    std : dict
        Std-part of surrogate evaluation
    true : dict
        Mean-part of model evaluation

    Returns
    -------
    dict
        Results of error measures

    """
    m = mean['Z']
    t = true['Z']
    diff = m-t
    return {'L2': np.linalg.norm(diff)}
    
    
if __name__=="__main__":
    random.seed(15)

    # Some model properties and settings
    t = np.linspace(0,10,10)
    ndim = 2
    
    # Plot some evaluations of the model 
    xx, yy = np.meshgrid(np.linspace(0,1,10),np.linspace(0,1,10))
    x = np.swapaxes(np.vstack([ xx.reshape(-1), yy.reshape(-1) ]),0,1)
    out_list = []
    for x_ in x:
        out_list.append(testfunction([x_],t)['Z'])
    out = np.array(out_list)
    mean_out = np.mean(out, axis = 0)
    std_out = np.std(out, axis = 0)
    
    for i in out:
        plt.plot(t,i,alpha = 0.2, color = 'blue')
    plt.show()
    plt.plot(t, mean_out)
    plt.show()
    plt.plot(t, std_out)
    plt.show()
    
    # Define the model
    Model = PyLinkForwardModel()
    Model.link_type = 'Function'
    Model.py_file = 'testfunction'
    Model.name = 'testfunction'
    Model.Output.names = ['Z']
    Model.func_args = {'t':t}

    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()
    priors = []
    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].name = "$\\theta_{"+str(i+1)+"}$"
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [0, 1]
        priors.append(np.random.uniform(0,1,1000))
    priors= np.swapaxes(np.array(priors),0,1)
    
    # Parameters for the convergence test
    mini = 2**ndim
    steps = [10,20,50]
    steps = [mini, 6, 8, 10, 15,20,25,30,35,40,50]
    
    # Get the samples and model evaluations
    train_samples = priors[:steps[-1]]
    test_samples = priors[steps[-1]:]
    test_data, test_std = Model.run_model_parallel(test_samples, mp = False)
    
#%% RUN THE SURROGATES ON THE FIRST BVR CODE

    # Set the surrogate parameters    
    MetaMod = sm.MetaModel(Inputs)
    MetaMod.meta_model_type = 'aPCE'
    MetaMod.pce_reg_method = 'FastARD'
    MetaMod.bootstrap_method = 'fast'
    MetaMod.n_bootstrap_itrs = 1
    MetaMod.pce_deg = np.arange(1,2)
    MetaMod.pce_q_norm = 0.85 if ndim < 5 else 0.5
    
    # Properties for Physics-informed training
    #MetaMod.LAR = True
    MetaMod.apply_constraints = False
    
    # Init ExpDesign - the rest of the ExpDesign is set in convergence_test
    ExpDesign = ExpDesigns(Inputs)
    
    # Start engine
    engine = Engine(MetaMod, Model, ExpDesign)

    # Run the convergence test
    results_FastArd = convergence_test(engine, steps, error, train_samples, test_samples, test_data)
    
    # Change the loss function base-type
    new_type = 'ols'
    engine.MetaModel.pce_reg_method = new_type
    results_OLS = convergence_test(engine, steps, error, train_samples, test_samples, test_data)
    
    # Do OLS with Constraints
    # The actual constraints can currently be found and changed in 
    # bayesvalidrox.surrogate_models.MetaModel.regression(). This interface
    # will be improved in future releases
    engine.MetaModel.apply_constraints = True
    engine.MetaModel.init_type = 'zeros'
    results_PI_0rand = convergence_test(engine, steps, error, train_samples, test_samples, test_data)
    
    # Plot the results
    for key in results_FastArd[1]:
        plt.plot(steps, results_OLS[1][key], linestyle = 'dashdot')
        plt.scatter(steps, results_OLS[1][key], label = 'OLS')
        plt.plot(steps, results_FastArd[1][key], linestyle = 'dashed')
        plt.scatter(steps, results_FastArd[1][key], label = 'FastARD')
        plt.plot(steps, results_PI_0rand[1][key], linestyle = 'dotted')
        plt.scatter(steps, results_PI_0rand[1][key], label = 'constrained')
        plt.legend()
        plt.title(f'Results for {key}')
        plt.show()
        
    # Get test statistics
    meanOLS, test_std = results_OLS[0][-1].eval_metamodel(test_samples)
    meanARD, test_std = results_FastArd[0][-1].eval_metamodel(test_samples)
    meanPI, test_std = results_PI_0rand[0][-1].eval_metamodel(test_samples)
    
    # Plot test statistics
    plt.plot(t, np.mean(meanOLS['Z'], axis = 0), label = 'OLS', linestyle = 'dashdot')
    plt.plot(t, np.mean(meanARD['Z'], axis = 0), label = 'FastARD', linestyle = 'dashed')
    plt.plot(t, np.mean(meanPI['Z'], axis = 0), label = 'constrained', linestyle = 'dotted')
    plt.plot(t, mean_out, label = 'data')
    plt.title('Approximation of mean')
    plt.legend()
    plt.show()
    plt.plot(t, np.std(meanOLS['Z'], axis = 0), label = 'OLS', linestyle = 'dashdot')
    plt.plot(t, np.std(meanARD['Z'], axis = 0), label = 'FastARD', linestyle = 'dashed')
    plt.plot(t, np.std(meanPI['Z'], axis = 0), label = 'constrained', linestyle = 'dotted')
    plt.plot(t, std_out, label = 'data')
    plt.title('Approximation of std')
    plt.legend()
    plt.show()
 
    