#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    pollution function. Here, the noise will be jointly inferred with the input
    parameters.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

"""

import numpy as np
import joblib

# import bayesvalidrox
# Add BayesValidRox path
import sys
sys.path.append("../../src/bayesvalidrox/")

from bayesvalidrox.pylink.pylink import PyLinkForwardModel
from bayesvalidrox.surrogate_models.inputs import Input
from bayesvalidrox.surrogate_models.exp_designs import ExpDesigns
from bayesvalidrox.surrogate_models.surrogate_models import MetaModel
#from bayesvalidrox.surrogate_models.meta_model_engine import MetaModelEngine
from bayesvalidrox.post_processing.post_processing import PostProcessing
from bayesvalidrox.bayes_inference.bayes_inference import BayesInference
from bayesvalidrox.bayes_inference.discrepancy import Discrepancy
from bayesvalidrox.surrogate_models.engine import Engine

if __name__ == "__main__":

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    # Define model options
    Model.link_type = 'Function'
    Model.py_file = 'pollution'
    Model.name = 'pollution'

    Model.Output.names = ['C']

    # Synthetic data for Bayesian inversion.
    # MAP = (10, 0.07, 1.0, 30.16)
    Model.observations = {}
    Model.observations['Time [s]'] = [40+5*i for i in range(21)]
    Model.observations['C'] = np.load("data/MeasuredData.npy")

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = [40+5*i for i in range(21)]
    Model.mc_reference['mean'] = np.load("data/mean.npy")
    Model.mc_reference['std'] = np.load("data/std.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    # mass of pollutant spilled at each location
    Inputs.add_marginals()
    Inputs.Marginals[0].name = 'M'
    Inputs.Marginals[0].dist_type = 'uniform'
    Inputs.Marginals[0].parameters = [7, 13]

    # diffusion rate in the channel
    Inputs.add_marginals()
    Inputs.Marginals[1].name = 'D'
    Inputs.Marginals[1].dist_type = 'uniform'
    Inputs.Marginals[1].parameters = [0.02, 0.12]

    # location of the second spill
    Inputs.add_marginals()
    Inputs.Marginals[2].name = 'L'
    Inputs.Marginals[2].dist_type = 'uniform'
    Inputs.Marginals[2].parameters = [0.01, 3]

    # time of the second spill
    Inputs.add_marginals()
    Inputs.Marginals[3].name = '$\\tau$'
    Inputs.Marginals[3].dist_type = 'uniform'
    Inputs.Marginals[3].parameters = [30.01, 30.295]

    # =====================================================
    # ======  POLYNOMIAL CHAOS EXPANSION METAMODELS  ======
    # =====================================================
    MetaModelOpts = MetaModel(Inputs)

    # Select if you want to preserve the spatial/temporal depencencies
    MetaModelOpts.dim_red_method = 'PCA'
    MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 5

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)BCS: Bayesian Compressive Sensing
    # 7)OMP: Orthogonal Matching Pursuit
    # 8)VBL: Variational Bayesian Learning
    # 9)EBL: Emperical Bayesian Learning
    MetaModelOpts.pce_reg_method = 'FastARD'

    # Bootstraping
    # 1) normal 2) fast
    MetaModelOpts.bootstrap_method = 'fast'
    MetaModelOpts.n_bootstrap_itrs = 100

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    MetaModelOpts.pce_deg = 6

    # q-quasi-norm 0<q<1 (default=1)
    # MetaModelOpts.pce_q_norm = 0.75

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------ Experimental Design --------
    ExpDesign = ExpDesigns(Inputs)

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    ExpDesign.method = 'normal'
    ExpDesign.n_init_samples = 150

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) korobov
    # 7) chebyshev(FT) 8) grid(FT) 9) nested_grid(FT) 10)user
    ExpDesign.sampling_method = 'latin_hypercube'

    # Provide the experimental design object with a hdf5 file
    # MetaModelOpts.ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Adaptive sparse arbitrary polynomial chaos expansion
    engine = Engine(MetaModelOpts, Model, ExpDesign)
    engine.start_engine()
    engine.train_normal()
    
    # Save PCE models
    with open(f"PCEModel_{Model.name}.pkl", "wb") as output:
        joblib.dump(engine.MetaModel, output, 2)

    # Load the objects
    # with open(f"PCEModel_{Model.name}.pkl", "rb") as input:
    #     PCEModel = joblib.load(input)

    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(engine)

    # Plot to check validation visually.
    PostPCE.valid_metamodel(n_samples=3)

    # Compute the moments and compare with the Monte-Carlo reference
    PostPCE.plot_moments()

    # Plot the sobol indices
    if MetaModelOpts.meta_model_type != 'GPE':
        total_sobol = PostPCE.sobol_indices()

    # Compute and print RMSE error
    valid_samples = np.load("data/validSet.npy")
    valid_out_dict = dict()
    valid_out_dict['Time [s]'] = [40+5*i for i in range(21)]
    valid_out_dict['C'] = np.load("data/validSetOutput.npy")
    PostPCE.check_accuracy(samples=valid_samples,
                           outputs=valid_out_dict)

    # =====================================================
    # ========  BAYESIAN INFERENCE (CALIBRATION) ==========
    # =====================================================
    BayesOpts = BayesInference(engine)
    BayesOpts.emulator = True

    # Select the inference method
    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters
    BayesOpts.mcmc_params = {
        'n_steps': 1e3,#5,
        'n_walkers': 30,
        'multiprocessing': False,
        'verbose': False
        }

    BayesOpts.plot_post_pred = True

    # ----- Define the discrepancy model -------
    # Let us infer the sigma2
    DiscOutputOpts = Input()
    # # OutputName = 'Z'
    DiscOutputOpts.add_marginals()
    DiscOutputOpts.Marginals[0].name = '$\\sigma^2_{\\epsilon}$'
    DiscOutputOpts.Marginals[0].dist_type = 'uniform'
    DiscOutputOpts.Marginals[0].parameters = [0.0, 0.1]
    BayesOpts.Discrepancy = Discrepancy(DiscOutputOpts)

    # Start the inference
    BayesCalib = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(BayesCalib, output, 2)

    # Load the objects
    # with open(f"Bayes_{Model.name}.pkl", "rb") as input:
    #     BayesCalib = joblib.load(input)

    # =====================================================
    # ===========  BAYESIAN MODEL VALIDATION ==============
    # =====================================================
    # ------- Model definition --------
    ValidModel = PyLinkForwardModel()

    # Define model options
    ValidModel.link_type = 'Function'
    ValidModel.py_file = 'pollution'
    ValidModel.func_args = {'s': np.array([6])}
    ValidModel.name = 'pollution-valid'

    ValidModel.Output.names = ['C']

    # Synthetic data for Bayesian inversion.
    # MAP = (10, 0.07, 1.0, 30.16)
    ValidModel.observations_valid = {}
    ValidModel.observations_valid['Time [s]'] = [40+5*i for i in range(21)]
    ValidModel.observations_valid['C'] = np.load("data/MeasuredData_valid.npy")

    # ------- Input definition --------
    ValidInputs = Input()

    ValidInputs.add_marginals()
    ValidInputs.Marginals[0].name = "M"
    ValidInputs.Marginals[0].input_data = BayesCalib.posterior_df['M']

    ValidInputs.add_marginals()
    ValidInputs.Marginals[1].name = "D"
    ValidInputs.Marginals[1].input_data = BayesCalib.posterior_df['D']

    ValidInputs.add_marginals()
    ValidInputs.Marginals[2].name = "L"
    ValidInputs.Marginals[2].input_data = BayesCalib.posterior_df['L']

    ValidInputs.add_marginals()
    ValidInputs.Marginals[3].name = "$\\tau$"
    ValidInputs.Marginals[3].input_data = BayesCalib.posterior_df['$\\tau$']


    import copy
    MetaModelOpts_valid = copy.deepcopy(MetaModelOpts)
    MetaModelOpts_valid.input_obj = ValidInputs
    valid_engine = Engine(MetaModelOpts_valid, ValidModel, copy.deepcopy(ExpDesign))
    valid_engine.start_engine()
    valid_engine.train_normal()

    # Save PCE models
    with open(f"PCEModel_{ValidModel.name}.pkl", "wb") as output:
        joblib.dump(valid_engine.MetaModel, output, 2)

    # Load the objects
    # with open(f"PCEModel_{ValidModel.name}.pkl", "rb") as input:
    #     ValidPCEModel = joblib.load(input)

    # ------- Baysian inference definition --------
    BayesOptsValid = BayesInference(valid_engine)

    BayesOptsValid.name = "Valid"
    BayesOptsValid.emulator = True

    # Bootstrap for BME and BF
    BayesOptsValid.bootstrap = True
    BayesOptsValid.n_bootstrap_itrs = 10

    # ----- Define the discrepancy model -------
    DiscrepancyOpts = Discrepancy("")
    DiscrepancyOpts.type = "Gaussian"
    DiscrepancyOpts.parameters = {
        "C": np.ones(21) * np.mean(
            BayesCalib.posterior_df['$\\sigma^2_{\\epsilon}$'].values)
    }
    BayesOptsValid.Discrepancy = DiscrepancyOpts

    # ----- Strat Bayesian inference -------
    BayesValid = BayesOptsValid.create_inference()

    #  Save class objects
    with open(f"Bayes_{ValidModel.name}.pkl", "wb") as output:
        joblib.dump(BayesValid, output, 2)
