#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug  9 16:25:43 2019

@author: farid
"""

import numpy as np
import os
import matplotlib.pyplot as plt
plt.style.use('seaborn-white')
plt.rcParams.update({'font.size': 16})
plt.rc('font', family='sans-serif', serif='Arial')
plt.rc('figure', figsize = (24, 16))
plt.rc('text', usetex=True)
plt.rc('xtick', labelsize=16)
plt.rc('ytick', labelsize=16)
plt.rc('axes', labelsize=16)


from PyLink.FuncForwardModel import FuncForwardModel
from surrogate_models.Input import Input
from surrogate_models.surrogate_models import Metamodel

if __name__ == "__main__":
    
    #=====================================================
    #=============   COMPUTATIONAL MODEL  ================
    #=====================================================
    Model = FuncForwardModel()
    
    Model.Type = 'Function'
    Model.pyFile = 'AnalyticalFunction'
    Model.Name = 'AnalyticFunc'
    
    Model.Output.Names = ['Z']
    

    #=====================================================
    #=========   PROBABILISTIC INPUT MODEL  ==============
    #=====================================================
    # Define the uncertain parameters with their mean and 
    # standard deviation
    Inputs = Input()
    
    Inputs.addMarginals()
    Inputs.Marginals[0].Name = 'x1'
    Inputs.Marginals[0].DistType = 'unif'
    Inputs.Marginals[0].Parameters =  [-5, 5]
    
    Inputs.addMarginals()
    Inputs.Marginals[1].Name = 'x2'
    Inputs.Marginals[1].DistType = 'unif'
    Inputs.Marginals[1].Parameters =  [-5, 5]

    
    #=====================================================
    #========  PCE METAMODELS with adaptive aPCE  ========
    #=====================================================    
    MetaModelOpts = Metamodel(Inputs)

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final 
    # metamodel.
    MetaModelOpts.MaxPceDegree = 10
    
    # Select the sparse least-square minimization method for 
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression 
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)SGDR: Stochastic gradient descent learning
    MetaModelOpts.RegMethod = 'OLS'
    
    # Print summary of the regression results
    #MetaModelOpts.DisplayFlag = True
    
    # ------ Experimental Design --------
    # Generate an experimental design of size NrExpDesign based on a latin 
    # hypercube sampling of the input model or user-defined values of X and/or Y:
    MetaModelOpts.addExpDesign()
    
    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.Method = 'normal'
    NrofInitSamples = 100
    MetaModelOpts.ExpDesign.NrSamples = NrofInitSamples
    
    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) chebyshev(FT) 
    # 7) korobov 8) grid(FT) 9) nested_grid(FT) 10)user
    MetaModelOpts.ExpDesign.SamplingMethod = 'latin_hypercube'
    
    # Sequential experimental design (needed only for sequential ExpDesign)
    MetaModelOpts.ExpDesign.MaxNSamples = 100
    MetaModelOpts.ExpDesign.ModifiedLOOThreshold = 1e-4
        
    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # -------- Exploration ------
    MetaModelOpts.ExpDesign.ExploreMethod = 'MC'
    MetaModelOpts.ExpDesign.NCandidate = 500
    MetaModelOpts.ExpDesign.NrofCandGroups = 4
    # -------- Optimality criteria: alphabetic ------
    MetaModelOpts.ExpDesign.ExploitMethod = 'alphabetic'
    
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    MetaModelOpts.ExpDesign.UtilityFunction = 'D-Opt'#['D-Opt', 'A-Opt', 'K-Opt']
    
    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Adaptive sparse arbitrary polynomial chaos expansion
    PCEModel_OLS = MetaModelOpts.create_metamodel(Model)
    
    # Extract the basis and coefficients
    Basis_OLS = PCEModel_OLS.BasisDict['Z']['y_2']
    COFFS_OLS = PCEModel_OLS.CoeffsDict['Z']['y_2']
    
    #=====================================================
    #===========  DEMO OF BAYESPCE METAMODEL  ===========
    #=====================================================
    fig1=plt.figure()
    barWidth = 0.3
    plt.title("Bar plot of the weights (OLS)")
     
    # Create cyan bars
    r = np.arange(len(Basis_OLS))
    plt.bar(r, COFFS_OLS, width = barWidth, color = 'cyan', bottom=0.001,
            edgecolor = 'black', capsize=7, label='Cofficients')
    
    # general layout
    plt.xticks([r + barWidth for r in range(len(r))], ['term'+str(i+1) for i in range(len(Basis_OLS))])

    plt.xlabel("Features")
    plt.ylabel("Values of the weights")
    plt.yscale('symlog', linthreshy=0.1) #('log')
    plt.hlines(0,xmin=r.min() , xmax=r.max() ,color='k')
    plt.xticks(rotation=45)
    plt.legend(loc="upper left")
    plt.show()
    
    # Make directory for saving the plots/outputs
    newpath = (r'Outputs_demo') 
    if not os.path.exists(newpath): os.makedirs(newpath)

    ## Save the plot the posterior (Model)
    fig1.savefig('./'+newpath+'/Coeffs_AaPCE.svg', bbox_inches='tight')
    plt.close()
    
    
    #=====================================================
    #===========  PCE METAMODELS with BRR aPCE  ==========
    #=====================================================  
    BRRMetaModelOpts = Metamodel(Inputs)
    BRRMetaModelOpts.MaxPceDegree = 10
    BRRMetaModelOpts.RegMethod = 'BRR'
    
    # ------ Experimental Design --------
    BRRMetaModelOpts.addExpDesign()
    BRRMetaModelOpts.ExpDesign.NrSamples = PCEModel_OLS.ExpDesign.X.shape[0]
    BRRMetaModelOpts.ExpDesign.SamplingMethod = 'user'
    BRRMetaModelOpts.ExpDesign.X = PCEModel_OLS.ExpDesign.X
    BRRMetaModelOpts.ExpDesign.Y = PCEModel_OLS.ExpDesign.Y
    BRRMetaModelOpts.ExpDesign.Method = 'normal'
    
    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Adaptive sparse arbitrary polynomial chaos expansion
    PCEModel_BRR = BRRMetaModelOpts.create_metamodel(Model)
    
    
    # Extract the basis and coefficients
    Basis_BRR = PCEModel_BRR.BasisDict['Z']['y_2']
    clf_poly = PCEModel_BRR.clf_poly['Z']['y_2']
   
    COFFS_BRR = clf_poly.coef_
    # Find the error(std) for coeffs
    COFFSError_BRR = clf_poly.sigma_.diagonal()

    #=====================================================
    #===========  DEMO OF BAYESPCE METAMODEL  ============
    #=====================================================
    fig2=plt.figure()
    barWidth = 0.3
    plt.title("Bar plot of the weights (BayesPCE)")
     
    # Create cyan bars
    r = np.arange(len(Basis_BRR))
    plt.bar(r, COFFS_BRR, width = barWidth, color = 'cyan', bottom=0.001,
            edgecolor = 'black', yerr=COFFSError_BRR, capsize=7, label='Cofficients')
    
    # general layout
    plt.xticks([r + barWidth for r in range(len(r))], ['term'+str(i+1) for i in range(len(Basis_BRR))])

    plt.xlabel("Features")
    plt.ylabel("Values of the weights")
    plt.yscale('symlog', linthreshy=0.1) #('log')
    plt.hlines(0,xmin=r.min() , xmax=r.max() ,color='k')
    plt.xticks(rotation=45)
    plt.legend(loc="upper left")
    plt.show()
    
    ## Save the plot the posterior (Model)
    fig2.savefig('./'+newpath+'/Coeffs_BRR.svg', bbox_inches='tight')
    plt.close()
    
    
